/*
 *	DEFS.H
 *	Tom Kerrigan's Simple Chess Program (TSCP)
 *
 *	Copyright 1997 Tom Kerrigan
 */


#define BOOL			int
#define TRUE			1
#define FALSE			0

#define MOVE_STACK		1120
#define HIST_STACK		400
#define UNDO_STACK		400

#define LIGHT			0
#define DARK			1

#define PAWN			0
#define KNIGHT			1
#define SILVER			2
#define LANCE			3
#define GOLD			4
#define KING			5

#define EMPTY			6
#define BOTH                    7

#define NUMSQUARES               64
#define LASTSQUARE               63
#define RANKS                     8
#define FILES                     8

/* useful squares */
#define A1				56
#define B1				57
#define C1				58
#define D1				59
#define E1				60
#define F1				61
#define G1				62
#define H1				63
#define A8				0
#define B8				1
#define C8				2
#define D8				3
#define E8				4
#define F8				5
#define G8				6
#define H8				7

#define GetFile(x)		(x & 7)
#define GetRank(x)		(x >> 3)


/* number of pieces for both sides since a bare king loses unless
   he can bare the other king in the next move in which case it's
   a draw.  */ 
extern int pieces_count[2];


/* This is the basic description of a move. promote is what
   piece to promote the pawn to, if the move is a pawn
   promotion. bits is a bitfield that describes the move,
   with the following bits:

   1	capture
   2	castle       ( unused
   4	en passant capture ( unused )
   8	pushing a pawn 2 squares
   16	pawn move
   32	promote

   It's union'ed with an integer so two moves can easily
   be compared with each other. */

typedef struct {
	char from;
	char to;
	char promote;
	char bits;
} move_bytes;

typedef union {
	move_bytes b;
	long u;
} move;

/* an element of the move stack. it's just a move with a
   score, so it can be sorted by the search functions. */
typedef struct {
	move m;
	long score;
} gen_t;

/* an element of the history stack, with the information
   necessary to take a move back. */
typedef struct {
	move m;
        int oldpiece;
	int capture;
} hist_t;


