/*
 *	EVAL.C
 *	Tom Kerrigan's Simple Chess Program (TSCP)
 *
 *	Copyright 1997 Tom Kerrigan
 */


#include <string.h>
#include "defs.h"
#include "data.h"
#include "protos.h"


#define KING_BEHIND_ENEMY_GOLD_PENALTY  -50
#define LANCE_SEMI_OPEN_FILE_BONUS	-10
#define LANCE_OPEN_FILE_BONUS		5
#define LANCE_ON_SEVENTH_BONUS		-20


/* the values of the pieces */
int piece_value[6] = {
  /* Pawn, Knight, Silver, Lance, Gold, King */
	100, 200, 350, 200, 400, 0
};

/* The "pcsq" arrays are piece/square tables. They're values
   added to the material value of the piece based on the
   location of the piece. */

int pawn_pcsq[64] = {
	  0,   0,   0,   0,   0,   0,   0,   0,
	  0,   0,   0,   0,   0,   0,   0,   0,
	  0,   0,   0,   0,   0,   0,   0,   0,
	  3,   6,  15,  15,  15,  15,   6,   3,
	  2,   2,   10,  12, 12,  12,   2,   4,
	 -5,   0,  -8,  -5,  -5,  -8,   0,  -5,
	  0,   0,   0,   0,   0,   0,   0,   0,
	  0,   0,   0,   0,   0,   0,   0,   0
};

int knight_pcsq[64] = {
	-10, -10, -10, -10, -10, -10, -10, -10,
	-10,   0,   0,   0,   0,   0,   0, -10,
	-10,   0,   0,   0,   0,   0,   0, -10,
	-10,   3,   0,   1,   1,  0,   3, -10,
	-10,   0,   0,   0,   0,   0,   0, -10,
	-10,   0,   2,   0,   0,   2,   0, -10,
	-10,   0,   0,   0,   0,   0,   0, -10,
	-10,   0,   0,   0,   0,   0,   0, -10
};

int lance_pcsq[64] = {
	  0,   0,   0,   0,   0,   0,   0,   0,
	-20,   0,   0,   0,   0,   0,   0, -20,
	-20,   0,   0,   0,   0,   0,   0, -20,
	-5,   0,   0,   0,   0,   0,   0, -5,
	-17,   0,   0,   0,   0,   0,   0, -17,
	-15,   0,   0,   0,   0,   0,   0, -15,
	-13,   0,   0,   0,   0,   0,   0, -13,
	-10,   0,   0,   0,   0,   0,   0, -10
};

int silver_pcsq[64] = {
	-10, -10, -10, -10, -10, -10, -10, -10,
	-10,   0,   0,   0,   0,   0,   0, -10,
	-10,   3,   5,   5,   5,   5,   3, -10,
	-10,   5,   10,  10,  10,  10,   5, -10,
	-10,   5,   10,  10,  10,   10,   5, -10,
	-10,   3,   5,   5,   5,   5,   3, -10,
	-10,   0,   3,   3,   3,   0,   0, -10,
	-10, -10, -20, -10, -10, -20, -10, -10
};

int gold_pcsq[64] = {
	-50, -40, -30, -25, -25, -30, -40, -50,
	-30,   0,   0,   0,   0,   0,   0, -30,
	-30,   3,   15,   15,   15,   15,   3, -30,
	-30,   5,   20,  25,  25,  20,  5, -30,
	-30,   5,   15,  20,  20,  15,  5, -30,
	-30,   10,   5,  15, 15,   5,   15,-30,
	-30,   0,   0,   3,   3,   0,   0, -30,
	-50, -40, -30, -25, -25, -30, -40, -50
};

int king_pcsq[64] = {
	  0,  1,  2,  3,  3,  2,  1,   0,
	 1,  2,  3,  4,  4,  3,  2,  1,
	 2,  3,  4,  5,  5,  4,  3,  2,
	 3,  4,  5,  6,  6,  5,  4,  3,
	 3,  4,  5,  6,  6,  5,  4,  3,
	 2,  3,  4,  5,  5,  4,  3,  2,
	 1,  2,  3,  4,  4,  3,  2,  1,
	  0,  1,  2,  3,  3,  2,  1,   0
};

int king_endgame_pcsq[64] = {
	  0,  10,  20,  30,  30,  20,  10,   0,
	 10,  20,  30,  40,  40,  30,  20,  10,
	 20,  30,  40,  50,  50,  40,  30,  20,
	 30,  40,  50,  60,  60,  50,  40,  30,
	 30,  40,  50,  60,  60,  50,  40,  30,
	 20,  30,  40,  50,  50,  40,  30,  20,
	 10,  20,  30,  40,  40,  30,  20,  10,
	  0,  10,  20,  30,  30,  20,  10,   0
};

/* The flip array is used to calculate the piece/square
   values for DARK pieces. The piece/square value of a
   LIGHT pawn is pawn_pcsq[sq] and the value of a DARK
   pawn is pawn_pcsq[flip[sq]] */
int flip[64] = {
	 56,  57,  58,  59,  60,  61,  62,  63,
	 48,  49,  50,  51,  52,  53,  54,  55,
	 40,  41,  42,  43,  44,  45,  46,  47,
	 32,  33,  34,  35,  36,  37,  38,  39,
	 24,  25,  26,  27,  28,  29,  30,  31,
	 16,  17,  18,  19,  20,  21,  22,  23,
	  8,   9,  10,  11,  12,  13,  14,  15,
	  0,   1,   2,   3,   4,   5,   6,   7
};


/* pawn_count[x][y] is the number of pawns of color x on file y.
   The array has two extra files (0 and 9) that will always be 0
   so we can use the same pawn evaluation code on every pawn without
   worrying about going out of bounds in the array. */
int pawn_count[2][10];

/* pawn_rank[x][y] is the rank of the least advanced pawn of color x
   on file y. If there isn't a pawn on that file, we assume it
   advanced off the board. This is a kludge that makes some of the
   pawn evaluation code work. This array has two extra files for the
   same reason pawn_count does. */
int pawn_rank[2][10];

int piece_mat[2];  /* the value of a side's pieces */
int pawn_mat[2];  /* the value of a side's pawns */

int eval()
{
	int i;
	int f;  /* file */
	int score[2];  /* each side's score */

	/* this is the first pass: set up pawn_count, pawn_rank,
	   piece_mat, and pawn_mat. */
	for (i = 0; i < 10; ++i) {
		pawn_count[LIGHT][i] = 0;
		pawn_count[DARK][i] = 0;
		pawn_rank[LIGHT][i] = -1;
		pawn_rank[DARK][i] = 8;
	}
	piece_mat[LIGHT] = 0;
	piece_mat[DARK] = 0;
	pawn_mat[LIGHT] = 0;
	pawn_mat[DARK] = 0;
	for (i = 0; i < 64; ++i) {
		if (color[i] == EMPTY)
			continue;
		if (piece[i] == PAWN) {
			pawn_mat[color[i]] += piece_value[PAWN];
			f = GetFile(i) + 1;  /* add 1 because of the extra file in the array */
			if (color[i] == LIGHT) {
				++pawn_count[LIGHT][f];
				if (pawn_rank[LIGHT][f] < GetRank(i))
					pawn_rank[LIGHT][f] = GetRank(i);
			}
			else {
				pawn_count[DARK][f]++;
				if (pawn_rank[DARK][f] > GetRank(i))
					pawn_rank[DARK][f] = GetRank(i);
			}
		}
		else
			piece_mat[color[i]] += piece_value[piece[i]];
	}

	/* this is the second pass: evaluate each piece */
	score[LIGHT] = piece_mat[LIGHT] + pawn_mat[LIGHT];
	score[DARK] = piece_mat[DARK] + pawn_mat[DARK];
	for (i = 0; i < 64; ++i) {
		if (color[i] == EMPTY)
			continue;
		if (color[i] == LIGHT) {
			switch (piece[i]) {
				case PAWN:
					score[LIGHT] += pawn_pcsq[i];
					break;
				case KNIGHT:
					score[LIGHT] += knight_pcsq[i];
					break;
				case SILVER:
					score[LIGHT] += silver_pcsq[i];
					break;
				case GOLD:
					score[LIGHT] += gold_pcsq[i];
					break;				       
				case LANCE:
				        score[LIGHT] += lance_pcsq[i];
					if (!pawn_count[LIGHT][GetFile(i) + 1]) {
						score[LIGHT] += LANCE_SEMI_OPEN_FILE_BONUS;
						if (!pawn_count[DARK][GetFile(i) + 1])
							score[LIGHT] += LANCE_OPEN_FILE_BONUS;
					}
					if (GetRank(i) == 1)
						score[LIGHT] += LANCE_ON_SEVENTH_BONUS;
					break;
				case KING:
				  /* if (piece_mat[DARK] <= 1200) */
				  score[LIGHT] += king_endgame_pcsq[i];
				  /*	else
					score[LIGHT] += eval_light_king(i); */
					break;
			}
		}
		else {
			switch (piece[i]) {
				case PAWN:
					score[DARK] += pawn_pcsq[flip[i]];
					break;
				case KNIGHT:
					score[DARK] += knight_pcsq[flip[i]];
					break;
				case SILVER:
					score[DARK] += silver_pcsq[flip[i]];
					break;
				case GOLD:
					score[DARK] += gold_pcsq[flip[i]];
					break;				       
				case LANCE:
				        score[DARK] += lance_pcsq[i];
					if (!pawn_count[DARK][GetFile(i) + 1]) {
						score[DARK] += LANCE_SEMI_OPEN_FILE_BONUS;
						if (!pawn_count[LIGHT][GetFile(i) + 1])
							score[DARK] += LANCE_OPEN_FILE_BONUS;
					}
					if (GetRank(i) == 6)
						score[DARK] += LANCE_ON_SEVENTH_BONUS;
					break;
				case KING:
					if (piece_mat[LIGHT] <= 1200)
						score[DARK] += king_endgame_pcsq[flip[i]];
					else
						score[DARK] += eval_dark_king(i);
					break;
			}
		}
	}

	/* the score[] array is set, now return the score relative
	   to the side to move */
	if (side == LIGHT)
		return score[LIGHT] - score[DARK];
	return score[DARK] - score[LIGHT];
}


/* The following two functions are the king safety functions.
   Here's the idea:
   First, the piece/square value is added to r (r is the value
     that's returned)
   Second, the pawns in front of the king are evaluated. If the
     king is on the queenside, the pawns on files a, b, and c
	 are evaluated. If the king is on the kingside, the pawns
	 on files f, g, and h are evaluated. The evaluation of each
	 file works like this: if there isn't a friendly pawn on the
	 file, it's considered open and a penalty is assessed. If
	 there is a friendly pawn on the file, and it hasn't been
	 pushed, nothing happens. If it's been pushed one square,
	 a penalty is assessed. If it's been pushed more than one
	 square, a penalty (usually bigger) is assessed. After this
	 is done for the three files, the enemy pawns are examined.
	 If they're getting too close to the king, it's a "pawn
	 storm" and penalties are assessed. If the king isn't on
	 the kingside or the queenside, i.e., if it's on one of the
	 two center files, then the three files closest to it are
	 examined. A penalty is assessed for each one of these files
	 that's empty.
   Finally, the score is multiplied by a scalar of the opponent's
     material. The rationale for this is that if there are problems
	 with your king safety and your opponent has a lot of pieces,
	 you're in trouble because it's easy for your opponent to
	 attack you. On the other hand, if you have problems with your
	 king safety and your opponent has no pieces, it doesn't really
	 matter, because he can't attack you. */

int eval_light_king(int sq)
{
	int r;  /* the value to return */
	int i;

	r = king_pcsq[sq];

	if (GetFile(sq) < 3) {
		
		/* a file */
		if (pawn_count[LIGHT][1]) {
			if (pawn_rank[LIGHT][1] == 6);
			else if (pawn_rank[LIGHT][1] == 5)
				r -= 2;
			else
				r -= 4;
		}
		else if (!pawn_count[DARK][1])
			r -= 8;

		/* b file */
		if (pawn_count[LIGHT][2]) {
			if (pawn_rank[LIGHT][2] == 6);
			else if (pawn_rank[LIGHT][2] == 5)
				r -= 2;
			else
				r -= 4;
		}
		else if (!pawn_count[DARK][2])
			r -= 8;

		/* c file */
		if (pawn_count[LIGHT][3]) {
			if (pawn_rank[LIGHT][3] == 6);
			else if (pawn_rank[LIGHT][3] ==5 )
				r -= 1;
			else
				r -= 3;
		}
		else if (!pawn_count[DARK][3])
			r -= 5;

		/* pawn storm */
		for (i = 1; i < 4; ++i) {
			if (!pawn_count[DARK][i])
				continue;
			if (pawn_rank[DARK][i] > 2) {
				if (pawn_rank[DARK][i] > 3)
					r -= 2;
				else
					r -= 1;
			}
		}
	}
	else if (GetFile(sq) > 4) {
		
		/* h file */
		if (pawn_count[LIGHT][8]) {
			if (pawn_rank[LIGHT][8] == 6);
			else if (pawn_rank[LIGHT][8] == 5)
				r -= 2;
			else
				r -= 4;
		}
		else if (!pawn_count[DARK][8])
			r -= 8;

		/* g file */
		if (pawn_count[LIGHT][7]) {
			if (pawn_rank[LIGHT][7] == 6);
			else if (pawn_rank[LIGHT][7] == 5)
				r -= 2;
			else
				r -= 4;
		}
		else if (!pawn_count[DARK][7])
			r -= 8;

		/* f file */
		if (pawn_count[LIGHT][6]) {
			if (pawn_rank[LIGHT][6] == 6);
			else if (pawn_rank[LIGHT][6] == 5)
				r -= 1;
			else
				r -= 3;
		}
		else if (!pawn_count[DARK][6])
			r -= 5;

		/* pawn storm */
		for (i = 6; i < 9; ++i) {
			if (!pawn_count[DARK][i])
				continue;
			if (pawn_rank[DARK][i] > 2) {
				if (pawn_rank[DARK][i] > 3)
					r -= 2;
				else
					r -= 1;
			}
		}
	}
	else {
		for (i = GetFile(sq); i <= GetFile(sq) + 2; ++i)
			if ((!pawn_count[LIGHT][i]) &&
					(!pawn_count[DARK][i]))
				r -= 2;
	}
	r *= piece_mat[DARK] / 600;
	return r;
}

int eval_dark_king(int sq)
{
	int r;  /* the value to return */
	int i;

	r = king_pcsq[flip[sq]];

	if (GetFile(sq) < 3) {
		
		/* a file */
		if (pawn_count[DARK][1]) {
			if (pawn_rank[DARK][1] == 1);
			else if (pawn_rank[DARK][1] == 2)
				r -= 2;
			else
				r -= 4;
		}
		else if (!pawn_count[LIGHT][1])
			r -= 8;

		/* b file */
		if (pawn_count[DARK][2]) {
			if (pawn_rank[DARK][2] == 1);
			else if (pawn_rank[DARK][2] == 2)
				r -= 2;
			else
				r -= 4;
		}
		else if (!pawn_count[LIGHT][2])
			r -= 8;

		/* c file */
		if (pawn_count[DARK][3]) {
			if (pawn_rank[DARK][3] == 1);
			else if (pawn_rank[DARK][3] == 2)
				r -= 1;
			else
				r -= 3;
		}
		else if (!pawn_count[LIGHT][3])
			r -= 5;

		/* pawn storm */
		for (i = 1;i<4;i++) {
			if (!pawn_count[LIGHT][i])
				continue;
			if (pawn_rank[LIGHT][i] < 5) {
				if (pawn_rank[LIGHT][i] < 4)
					r -= 2;
				else
					r -= 1;
			}
		}
	}
	else if (GetFile(sq) > 4) {
		
		/* h file */
		if (pawn_count[DARK][8]) {
			if (pawn_rank[DARK][8] == 1);
			else if (pawn_rank[DARK][8] == 2)
				r -= 2;
			else
				r -= 4;
		}
		else if (!pawn_count[LIGHT][8])
			r -= 8;

		/* g file */
		if (pawn_count[DARK][7]) {
			if (pawn_rank[DARK][7] == 1);
			else if (pawn_rank[DARK][7] == 2)
				r -= 2;
			else
				r -= 4;
		}
		else if (!pawn_count[LIGHT][7])
			r -= 8;

		/* f file */
		if (pawn_count[DARK][6]) {
			if (pawn_rank[DARK][6] == 1);
			else if (pawn_rank[DARK][6] == 2)
				r -= 1;
			else
				r -= 3;
		}
		else if (!pawn_count[LIGHT][6])
			r -= 5;

		/* pawn storm */
		for (i = 6; i < 9; ++i) {
			if (!pawn_count[LIGHT][i])
				continue;
			if (pawn_rank[LIGHT][i] < 5) {
				if (pawn_rank[LIGHT][i] < 4)
					r -= 2;
				else
					r -= 1;
			}
		}
	}
	else {
		for (i = GetFile(sq); i <= GetFile(sq) + 2; ++i)
			if ((!pawn_count[LIGHT][i]) &&
					(!pawn_count[DARK][i]))
				r -= 2;
	}
	r *= piece_mat[LIGHT] / 600;
	return r;
}






