/*
 *	MAIN.C
 *      E.Werner
 *      derived from
 *	Tom Kerrigan's Simple Chess Program (TSCP)
 *
 *	Copyright 1997 Tom Kerrigan
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include "defs.h"
#include "data.h"
#include "protos.h"

char *side_name[2] = { "Sente", "Gote" };

/* get_ms() returns the milliseconds elapsed since midnight,
   January 1, 1970. */

#include <sys/timeb.h>
long get_ms()
{
	struct timeb timebuffer;
	ftime(&timebuffer);
	return (timebuffer.time * 1000) + timebuffer.millitm;
}


/* main() is basically an infinite loop that either calls
   think() when it's the computer's turn to move or prompts
   the user for a command (and deciphers it). */

void main(int argc, const char *argv[])
{
	int computer_side;
	char s[256];
	int from, to, i;
	BOOL found, promote;

	printf("\n");
	printf("Heian Shogi 0.01 by E.Werner\n");
	printf("   derived from\n");
	printf("Tom Kerrigan's Simple Chess Program (TSCP)\n");
	printf("version 1.52, 6/29/00\n");
	printf("\n");
	printf("\"help\" displays a list of commands.\n");
	printf("\n");
	init();
	gen();
	computer_side = EMPTY;
	max_time = 1000000;
	max_depth = 4;
	for (;;) {
		if ((side == computer_side)|| (computer_side == BOTH )) {  /* computer's turn */
			
			/* think about the move and make it */
			think(FALSE);
			if ( !pv[0][0].u ) {
				printf("%s loses (no legal moves or bare king)\n", side_name[side] );
				computer_side = EMPTY;
				continue;
			}
			printf("Computer's move: %s\n", move_str(pv[0][0].b));
			makemove(pv[0][0].b);
			if ( computer_side == BOTH ) {
			  print_board();
			}

			/* save the "backup" data because it will be overwritten */
			undo_dat[undos] = hist_dat[0];
			++undos;

			ply = 0;
			gen();
			continue;
		}

		/* get user input */
		printf("%s> ", argv[0]);
		scanf("%s", s);
		if (!strcmp(s, "auto")) {
		        computer_side = BOTH;
			continue;
		}
		if (!strcmp(s, "save")) {
		        save_game();
			continue;
		}
		if (!strcmp(s, "load")) {
		        load_game();
			continue;
		}
		if (!strcmp(s, "on")) {
			computer_side = side;
			continue;
		}
		if (!strcmp(s, "off")) {
			computer_side = EMPTY;
			continue;
		}
		if (!strcmp(s, "st")) {
			scanf("%d", &max_time);
			max_time *= 1000;
			max_depth = 10000;
			continue;
		}
		if (!strcmp(s, "sd")) {
			scanf("%d", &max_depth);
			max_time = 1000000;
			continue;
		}
		if (!strcmp(s, "undo")) {
			if (!undos)
				continue;
			computer_side = EMPTY;
			--undos;
			hist_dat[0] = undo_dat[undos];
			ply = 1;
			takeback();
			gen();
			continue;
		}
		if (!strcmp(s, "new")) {
			computer_side = EMPTY;
			init();
			gen();
			continue;
		}
		if (!strcmp(s, "d")) {
			print_board();
			continue;
		}
		if (!strcmp(s, "bye") || !strcmp(s, "quit")) {
			printf("Share and enjoy!\n");
			exit(0);
		}
		if (!strcmp(s, "xboard")) {
			xboard();
			exit(0);
		}
		if (!strcmp(s, "hint")) {
			think(TRUE);
			if (!pv[0][0].u)
				continue;
			printf("Hint: %s\n", move_str(pv[0][0].b));
			fflush(stdout);
			continue;
		}
		if (!strcmp(s, "help")) {
			printf("auto - computer plays both sides\n");		  
			printf("on - computer plays for the side to move\n");
			printf("off - computer stops playing\n");
			printf("st n - search for n seconds per move\n");
			printf("sd n - search n ply per move\n");
			printf("undo - takes back a move\n");
			printf("new - starts a new game\n");
			printf("d - display the board\n");
			printf("load - load a game file\n");
			printf("save - save the current game to a file\n");
			printf("bye/quit  - exit the program\n");
			printf("xboard - switch to XBoard mode\n");
			printf("Enter moves in coordinate notation, e.g., 2f2e, 3e3f+\n");
			continue;
		}

		/* maybe the user entered a move? */
		from = 56 - s[0];
		from += 8 * (s[1] - 'a');
		to = 56 - s[2];
		to += 8 * (s[3] - 'a');
		if ( strlen(s) > 4 && s[4] == '+' ) {
		  promote = TRUE;
		} else promote = FALSE;

		/* loop through the moves to see if it's legal */
		found = FALSE;
		for (i = 0; i < gen_end[ply]; ++i) {
		  if ( promote && ! ( gen_dat[i].m.b.bits & 32 )) 
		    continue;
		  if  ( !promote && (gen_dat[i].m.b.bits & 32)) 
		    continue;
		  if (gen_dat[i].m.b.from == from && gen_dat[i].m.b.to == to) {
		    found = TRUE;
		    break;
		  }
		}
		if (!found || !makemove(gen_dat[i].m.b))
			printf("Illegal move.\n");
		else {
			/* save the "backup" data because it will be overwritten */
			undo_dat[undos] = hist_dat[0];
			++undos;

			ply = 0;
			gen();
		}

	}
}


/* move_str returns a string with move m in coordinate notation */

char *move_str(move_bytes m)
{
	static char str[6];

	if (m.bits & 32)  /* promotion */
		sprintf(str, "%d%c%d%c+",
				8 - GetFile(m.from),
				GetRank(m.from) + 'a',
				8 - GetFile(m.to),
				GetRank(m.to) + 'a');
	else
		sprintf(str, "%d%c%d%c",
				8 - GetFile(m.from),
				GetRank(m.from) + 'a',
				8 - GetFile(m.to),
				GetRank(m.to) + 'a');

	return str;
}


/* print_board prints the board */

void print_board()
{
	int i;
	
	printf("\na ");
	for (i = 0; i < NUMSQUARES; ++i) {
		switch (color[i]) {
			case EMPTY:
				printf(" .");
				break;
			case LIGHT:
				printf(" %c", piece_char[piece[i]]);
				break;
			case DARK:
				printf(" %c", piece_char[piece[i]] + ('a' - 'A'));
				break;
		}
		if ((i + 1) % RANKS == 0 && i != LASTSQUARE)
			printf("\n%c ", 'b' + GetRank(i));
	}
	printf("\n\n   8 7 6 5 4 3 2 1\n\n");
}


/* xboard() is a substitute for main() that is XBoard
   and WinBoard compatible. See the following page for details:
   http://www.research.digital.com/SRC/personal/mann/xboard/engine-intf.html */

void xboard()
{
	int computer_side;
	char line[256], command[256];
	int from, to, i;
	BOOL found;

	signal(SIGINT, SIG_IGN);
	printf("\n");
	init();
	gen();
	computer_side = EMPTY;
	for (;;) {
		if (side == computer_side) {
			think(TRUE);
			if (!pv[0][0].u) {
				computer_side = EMPTY;
				continue;
			}
			printf("move %s\n", move_str(pv[0][0].b));
			fflush(stdout);
			makemove(pv[0][0].b);
			undo_dat[undos] = hist_dat[0];
			++undos;
			ply = 0;
			gen();
			continue;
		}
		fgets(line, 256, stdin);
		if (line[0] == '\n')
			continue;
		sscanf(line, "%s", command);
		if (!strcmp(command, "xboard"))
			continue;
		if (!strcmp(command, "new")) {
			init();
			gen();
			computer_side = DARK;
			continue;
		}
		if (!strcmp(command, "quit"))
			return;
		if (!strcmp(command, "force")) {
			computer_side = EMPTY;
			continue;
		}
		if (!strcmp(command, "white")) {
			side = LIGHT;
			xside = DARK;
			undos = 0;
			ply = 0;
			gen();
			computer_side = DARK;
			continue;
		}
		if (!strcmp(command, "black")) {
			side = DARK;
			xside = LIGHT;
			undos = 0;
			ply = 0;
			gen();
			computer_side = LIGHT;
			continue;
		}
		if (!strcmp(command, "st")) {
			sscanf(line, "st %d", &max_time);
			max_time *= 1000;
			max_depth = 10000;
			continue;
		}
		if (!strcmp(command, "sd")) {
			sscanf(line, "sd %d", &max_depth);
			max_time = 1000000;
			continue;
		}
		if (!strcmp(command, "time")) {
			sscanf(line, "time %d", &max_time);
			max_time *= 10;
			max_time /= 30;
			max_depth = 10000;
			continue;
		}
		if (!strcmp(command, "otim")) {
			continue;
		}
		if (!strcmp(command, "go")) {
			computer_side = side;
			continue;
		}
		if (!strcmp(command, "hint")) {
			think(TRUE);
			if (!pv[0][0].u)
				continue;
			printf("Hint: %s\n", move_str(pv[0][0].b));
			fflush(stdout);
			continue;
		}
		if (!strcmp(command, "undo")) {
			if (!undos)
				continue;
			--undos;
			hist_dat[0] = undo_dat[undos];
			ply = 1;
			takeback();
			gen();
			continue;
		}
		if (!strcmp(command, "remove")) {
			if (undos < 2)
				continue;
			--undos;
			hist_dat[0] = undo_dat[undos];
			ply = 1;
			takeback();
			--undos;
			hist_dat[0] = undo_dat[undos];
			ply = 1;
			takeback();
			gen();
			continue;
		}
		from =  8 - line[0] ;
		printf("%s\n",from);
		from += 8 * (8 - (line[1] - '0'));
		printf("%s\n",from);
		to = 8  - line[2] ;
		printf("%s\n",to);
		to += 8 * (8 - (line[3] - '0'));
		printf("%s\n",to);
		found = FALSE;

		/* promotion stuff */
		for (i = 0; i < gen_end[ply]; ++i)
			if (gen_dat[i].m.b.from == from && gen_dat[i].m.b.to == to) {
				found = TRUE;
				if (gen_dat[i].m.b.bits & 32)
					switch (line[4]) {
						case 'n':
							break;
						case 'b':
							i += 1;
							break;
						case 'r':
							i += 2;
							break;
						default:
							i += 3;
							break;
					}
				break;
			}
		if (!found || !makemove(gen_dat[i].m.b)) {
			printf("Error (unknown command): %s\n", command);
			fflush(stdout);
		}
		else {
			undo_dat[undos] = hist_dat[0];
			++undos;
			ply = 0;
			gen();
		}
	}
}

void load_game( void ) {
  char fn[1024], s[5];
  int from, to, i;
  BOOL promote, found;
  FILE *fh;
  printf("Enter load file name: ");
  scanf("%s",fn);
  fh = fopen(fn,"rb");
  if (! fh) {
    printf("No such file!\n");
    return;
  }
  do {
    fscanf(fh,"%s\n",s);
    from = 56 - s[0];
    from += 8 * (s[1] - 'a');
    to = 56 - s[2];
    to += 8 * (s[3] - 'a');
    if ( strlen(s) > 4 && s[4] == '+' ) {
      promote = TRUE;
    } else promote = FALSE;
    
    /* loop through the moves to see if it's legal */
    found = FALSE;
    for (i = 0; i < gen_end[ply]; ++i) {
      if ( promote && ! ( gen_dat[i].m.b.bits & 32 )) 
	continue;
      if  ( !promote && (gen_dat[i].m.b.bits & 32)) 
	continue;
      if (gen_dat[i].m.b.from == from && gen_dat[i].m.b.to == to) {
	found = TRUE;
	break;
      }
    }
    if (!found || !makemove(gen_dat[i].m.b))
      printf("Illegal move.\n");
    else {
      /* save the "backup" data because it will be overwritten */
      undo_dat[undos] = hist_dat[0];
      ++undos;
      
      ply = 0;
      gen();
    }

  } while (!feof(fh));
  printf("%d plies\n", undos );
  fclose(fh);
}


void save_game( void ) {
  char fn[1024], s[5];
  int i;
  FILE *fh;
  printf("Enter save file name: ");
  scanf("%s",fn);
  fh = fopen(fn,"rb");
  if (fh) {
    printf("File already exists!\n");
    fclose(fh);
    return;
  }
  fh = fopen(fn,"wb");
  printf("%d\n",undos);
  for (i=0;i < undos; i++) {
    fprintf(fh,"%s\n",move_str(undo_dat[i].m.b));
  }
  fclose(fh);
}

